@@ -23,7 +23,7 @@ _username = None
 _password = None
 _max_connections = 10
 
-def setup(hosts, username=None, password=None, max_connections=10):
+def setup(hosts, username=None, password=None, max_connections=10, default_keyspace=None, lazy=False):
     """
     Records the hosts and connects to one of them
 
@@ -36,7 +36,11 @@ def setup(hosts, username=None, password=None, max_connections=10):
     _username = username
     _password = password
     _max_connections = max_connections
-    
+
+    if default_keyspace:
+        from cqlengine import models
+        models.DEFAULT_KEYSPACE = default_keyspace
+
     for host in hosts:
         host = host.strip()
         host = host.split(':')
@@ -51,9 +55,10 @@ def setup(hosts, username=None, password=None, max_connections=10):
         raise CQLConnectionError("At least one host required")
 
     random.shuffle(_hosts)
-    
-    con = ConnectionPool.get()
-    ConnectionPool.put(con)
+
+    if not lazy:
+        con = ConnectionPool.get()
+        ConnectionPool.put(con)
 
 
 class ConnectionPool(object):
@@ -73,7 +78,7 @@ class ConnectionPool(object):
                 cls._queue.get().close()
         except:
             pass
-    
+
     @classmethod
     def get(cls):
         """
@@ -119,7 +124,7 @@ class ConnectionPool(object):
         global _hosts
         global _username
         global _password
-        
+
         if not _hosts:
             raise CQLConnectionError("At least one host required")
 
@@ -128,7 +133,7 @@ class ConnectionPool(object):
         new_conn = cql.connect(host.name, host.port, user=_username, password=_password)
         new_conn.set_cql_version('3.0.0')
         return new_conn
-        
+
 
 class connection_manager(object):
     """
@@ -140,7 +145,7 @@ class connection_manager(object):
         self.keyspace = None
         self.con = ConnectionPool.get()
         self.cur = None
-        
+
     def close(self):
         if self.cur: self.cur.close()
         ConnectionPool.put(self.con)
@@ -165,10 +170,11 @@ class connection_manager(object):
                 self.cur = self.con.cursor()
                 self.cur.execute(query, params)
                 return self.cur
+            except cql.ProgrammingError as ex:
+                raise CQLEngineException(unicode(ex))
             except TTransportException:
                 #TODO: check for other errors raised in the event of a connection / server problem
                 #move to the next connection and set the connection pool
-                self.con = None
                 _host_idx += 1
                 _host_idx %= len(_hosts)
                 self.con.close()
