@@ -1,10 +1,14 @@
 from collections import namedtuple
 import copy
+from datetime import datetime
 from hashlib import md5
 from time import time
+from uuid import uuid1
+from cqlengine import BaseContainerColumn, BaseValueManager, Map
 
 from cqlengine.connection import connection_manager
 from cqlengine.exceptions import CQLEngineException
+from cqlengine.functions import BaseQueryFunction
 
 #CQL 3 reference:
 #http://www.datastax.com/docs/1.1/references/cql/index
@@ -27,7 +31,7 @@ class QueryOperator(object):
         #the identifier is a unique key that will be used in string
         #replacement on query strings, it's created from a hash
         #of this object's id and the time
-        self.identifier = md5(str(id(self)) + str(time())).hexdigest()
+        self.identifier = uuid1().hex
 
         #perform validation on this operator
         self.validate_operator()
@@ -39,7 +43,10 @@ class QueryOperator(object):
         Returns this operator's portion of the WHERE clause
         :param valname: the dict key that this operator's compare value will be found in
         """
-        return '{} {} :{}'.format(self.column.db_field_name, self.cql_symbol, self.identifier)
+        if isinstance(self.value, BaseQueryFunction):
+            return '"{}" {} {}'.format(self.column.db_field_name, self.cql_symbol, self.value.to_cql(self.identifier))
+        else:
+            return '"{}" {} :{}'.format(self.column.db_field_name, self.cql_symbol, self.identifier)
 
     def validate_operator(self):
         """
@@ -74,7 +81,10 @@ class QueryOperator(object):
         this should return the dict: {'colval':<self.value>}
         SELECT * FROM column_family WHERE colname=:colval
         """
-        return {self.identifier: self.value}
+        if isinstance(self.value, BaseQueryFunction):
+            return {self.identifier: self.column.to_database(self.value.get_value())}
+        else:
+            return {self.identifier: self.column.to_database(self.value)}
 
     @classmethod
     def get_operator(cls, symbol):
@@ -100,6 +110,27 @@ class InOperator(EqualsOperator):
     symbol = 'IN'
     cql_symbol = 'IN'
 
+    class Quoter(object):
+        """
+        contains a single value, which will quote itself for CQL insertion statements
+        """
+        def __init__(self, value):
+            self.value = value
+
+        def __str__(self):
+            from cql.query import cql_quote as cq
+            return '(' + ', '.join([cq(v) for v in self.value]) + ')'
+
+    def get_dict(self):
+        if isinstance(self.value, BaseQueryFunction):
+            return {self.identifier: self.column.to_database(self.value.get_value())}
+        else:
+            try:
+                values = [v for v in self.value]
+            except TypeError:
+                raise QueryException("in operator arguments must be iterable, {} found".format(self.value))
+            return {self.identifier: self.Quoter([self.column.to_database(v) for v in self.value])}
+
 class GreaterThanOperator(QueryOperator):
     symbol = "GT"
     cql_symbol = '>'
@@ -116,18 +147,72 @@ class LessThanOrEqualOperator(QueryOperator):
     symbol = "LTE"
     cql_symbol = '<='
 
+class BatchType(object):
+    Unlogged    = 'UNLOGGED'
+    Counter     = 'COUNTER'
+
+class BatchQuery(object):
+    """
+    Handles the batching of queries
+
+    http://www.datastax.com/docs/1.2/cql_cli/cql/BATCH
+    """
+
+    def __init__(self, batch_type=None, timestamp=None):
+        self.queries = []
+        self.batch_type = batch_type
+        if timestamp is not None and not isinstance(timestamp, datetime):
+            raise CQLEngineException('timestamp object must be an instance of datetime')
+        self.timestamp = timestamp
+
+    def add_query(self, query, params):
+        self.queries.append((query, params))
+
+    def execute(self):
+        if len(self.queries) == 0:
+            # Empty batch is a no-op
+            return
+
+        opener = 'BEGIN ' + (self.batch_type + ' ' if self.batch_type else '') + ' BATCH'
+        if self.timestamp:
+            epoch = datetime(1970, 1, 1)
+            ts = long((self.timestamp - epoch).total_seconds() * 1000)
+            opener += ' USING TIMESTAMP {}'.format(ts)
+
+        query_list = [opener]
+        parameters = {}
+        for query, params in self.queries:
+            query_list.append('  ' + query)
+            parameters.update(params)
+
+        query_list.append('APPLY BATCH;')
+
+        with connection_manager() as con:
+            con.execute('\n'.join(query_list), parameters)
+
+        self.queries = []
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        #don't execute if there was an exception
+        if exc_type is not None: return
+        self.execute()
+
 class QuerySet(object):
 
     def __init__(self, model):
         super(QuerySet, self).__init__()
         self.model = model
-        self.column_family_name = self.model.column_family_name()
 
         #Where clause filters
         self._where = []
 
         #ordering arguments
-        self._order = None
+        self._order = []
+
+        self._allow_filtering = False
 
         #CQL has a default limit of 10000, it's defined here
         #because explicit is better than implicit
@@ -143,6 +228,12 @@ class QuerySet(object):
         self._result_cache = None
         self._result_idx = None
 
+        self._batch = None
+
+    @property
+    def column_family_name(self):
+        return self.model.column_family_name()
+
     def __unicode__(self):
         return self._select_query()
 
@@ -157,6 +248,12 @@ class QuerySet(object):
         for k,v in self.__dict__.items():
             if k in ['_con', '_cur', '_result_cache', '_result_idx']:
                 clone.__dict__[k] = None
+            elif k == '_batch':
+                # we need to keep the same batch instance across
+                # all queryset clones, otherwise the batched queries
+                # fly off into other batch instances which are never
+                # executed, thx @dokai
+                clone.__dict__[k] = self._batch
             else:
                 clone.__dict__[k] = copy.deepcopy(v, memo)
 
@@ -164,7 +261,7 @@ class QuerySet(object):
 
     def __len__(self):
         return self.count()
-    
+
     def __del__(self):
         if self._con:
             self._con.close()
@@ -180,6 +277,14 @@ class QuerySet(object):
         equal_ops = [w for w in self._where if isinstance(w, EqualsOperator)]
         if not any([w.column.primary_key or w.column.index for w in equal_ops]):
             raise QueryException('Where clauses require either a "=" or "IN" comparison with either a primary key or indexed field')
+
+        if not self._allow_filtering:
+            #if the query is not on an indexed field
+            if not any([w.column.index for w in equal_ops]):
+                if not any([w.column._partition_key for w in equal_ops]):
+                    raise QueryException('Filtering on a clustering key without a partition key is not allowed unless allow_filtering() is called on the querset')
+
+
         #TODO: abuse this to see if we can get cql to raise an exception
 
     def _where_clause(self):
@@ -205,23 +310,28 @@ class QuerySet(object):
             fields = [f for f in fields if f in self._only_fields]
         db_fields = [self.model._columns[f].db_field_name for f in fields]
 
-        qs = ['SELECT {}'.format(', '.join(db_fields))]
+        qs = ['SELECT {}'.format(', '.join(['"{}"'.format(f) for f in db_fields]))]
         qs += ['FROM {}'.format(self.column_family_name)]
 
         if self._where:
             qs += ['WHERE {}'.format(self._where_clause())]
 
         if self._order:
-            qs += ['ORDER BY {}'.format(self._order)]
+            qs += ['ORDER BY {}'.format(', '.join(self._order))]
 
         if self._limit:
             qs += ['LIMIT {}'.format(self._limit)]
 
+        if self._allow_filtering:
+            qs += ['ALLOW FILTERING']
+
         return ' '.join(qs)
 
     #----Reads------
 
     def _execute_query(self):
+        if self._batch:
+            raise CQLEngineException("Only inserts, updates, and deletes are available in batch mode")
         if self._result_cache is None:
             self._con = connection_manager()
             self._cur = self._con.execute(self._select_query(), self._where_values())
@@ -241,7 +351,7 @@ class QuerySet(object):
                 value_dict = dict(zip(names, values))
                 self._result_idx += 1
                 self._result_cache[self._result_idx] = self._construct_instance(value_dict)
-                
+
             #return the connection to the connection pool if we have all objects
             if self._result_cache and self._result_cache[-1] is not None:
                 self._con.close()
@@ -283,7 +393,7 @@ class QuerySet(object):
             else:
                 self._fill_result_cache_to_idx(s)
                 return self._result_cache[s]
-            
+
 
     def _construct_instance(self, values):
         #translate column names to model names
@@ -294,7 +404,21 @@ class QuerySet(object):
                 field_dict[db_map[key]] = val
             else:
                 field_dict[key] = val
-        return self.model(**field_dict)
+        instance = self.model(**field_dict)
+        instance._is_persisted = True
+        return instance
+
+    def batch(self, batch_obj):
+        """
+        Adds a batch query to the mix
+        :param batch_obj:
+        :return:
+        """
+        if not isinstance(batch_obj, BatchQuery):
+            raise CQLEngineException('batch_obj must be a BatchQuery instance')
+        clone = copy.deepcopy(self)
+        clone._batch = batch_obj
+        return clone
 
     def first(self):
         try:
@@ -356,52 +480,59 @@ class QuerySet(object):
                     '{} objects found'.format(len(self._result_cache)))
         else:
             return self[0]
-        
 
-
-    def order_by(self, colname):
+    def order_by(self, *colnames):
         """
         orders the result set.
-        ordering can only select one column, and it must be the second column in a composite primary key
+        ordering can only use clustering columns.
 
         Default order is ascending, prepend a '-' to the column name for descending
         """
-        if colname is None:
+        if len(colnames) == 0:
             clone = copy.deepcopy(self)
-            clone._order = None
+            clone._order = []
             return clone
 
-        order_type = 'DESC' if colname.startswith('-') else 'ASC'
-        colname = colname.replace('-', '')
+        conditions = []
+        for colname in colnames:
+            order_type = 'DESC' if colname.startswith('-') else 'ASC'
+            colname = colname.replace('-', '')
 
-        column = self.model._columns.get(colname)
-        if column is None:
-            raise QueryException("Can't resolve the column name: '{}'".format(colname))
+            column = self.model._columns.get(colname)
+            if column is None:
+                raise QueryException("Can't resolve the column name: '{}'".format(colname))
 
-        #validate the column selection
-        if not column.primary_key:
-            raise QueryException(
-                "Can't order on '{}', can only order on (clustered) primary keys".format(colname))
+            #validate the column selection
+            if not column.primary_key:
+                raise QueryException(
+                    "Can't order on '{}', can only order on (clustered) primary keys".format(colname))
 
-        pks = [v for k,v in self.model._columns.items() if v.primary_key]
-        if column == pks[0]:
-            raise QueryException(
-                "Can't order by the first primary key, clustering (secondary) keys only")
+            pks = [v for k, v in self.model._columns.items() if v.primary_key]
+            if column == pks[0]:
+                raise QueryException(
+                    "Can't order by the first primary key (partition key), clustering (secondary) keys only")
+
+            conditions.append('"{}" {}'.format(column.db_field_name, order_type))
 
         clone = copy.deepcopy(self)
-        clone._order = '{} {}'.format(column.db_field_name, order_type)
+        clone._order.extend(conditions)
         return clone
 
     def count(self):
         """ Returns the number of rows matched by this query """
+        if self._batch:
+            raise CQLEngineException("Only inserts, updates, and deletes are available in batch mode")
         #TODO: check for previous query execution and return row count if it exists
         if self._result_cache is None:
             qs = ['SELECT COUNT(*)']
             qs += ['FROM {}'.format(self.column_family_name)]
             if self._where:
                 qs += ['WHERE {}'.format(self._where_clause())]
+            if self._allow_filtering:
+                qs += ['ALLOW FILTERING']
+
             qs = ' '.join(qs)
-    
+
             with connection_manager() as con:
                 cur = con.execute(qs, self._where_values())
                 return cur.fetchone()[0]
@@ -410,7 +541,7 @@ class QuerySet(object):
 
     def limit(self, v):
         """
-        Sets the limit on the number of results returned 
+        Sets the limit on the number of results returned
         CQL has a default limit of 10,000
         """
         if not (v is None or isinstance(v, (int, long))):
@@ -425,6 +556,15 @@ class QuerySet(object):
         clone._limit = v
         return clone
 
+    def allow_filtering(self):
+        """
+        Enables the unwise practive of querying on a clustering
+        key without also defining a partition key
+        """
+        clone = copy.deepcopy(self)
+        clone._allow_filtering = True
+        return clone
+
     def _only_or_defer(self, action, fields):
         clone = copy.deepcopy(self)
         if clone._defer_fields or clone._only_fields:
@@ -454,19 +594,64 @@ class QuerySet(object):
         """ Don't load these fields for the returned query """
         return self._only_or_defer('defer', fields)
 
-    #----writes----
-    def save(self, instance):
+    def create(self, **kwargs):
+        return self.model(**kwargs).batch(self._batch).save()
+
+    #----delete---
+    def delete(self, columns=[]):
+        """
+        Deletes the contents of a query
+        """
+        #validate where clause
+        partition_key = self.model._primary_keys.values()[0]
+        if not any([c.column.db_field_name == partition_key.db_field_name for c in self._where]):
+            raise QueryException("The partition key must be defined on delete queries")
+        qs = ['DELETE FROM {}'.format(self.column_family_name)]
+        qs += ['WHERE {}'.format(self._where_clause())]
+        qs = ' '.join(qs)
+
+        if self._batch:
+            self._batch.add_query(qs, self._where_values())
+        else:
+            with connection_manager() as con:
+                con.execute(qs, self._where_values())
+
+class DMLQuery(object):
+    """
+    A query object used for queries performing inserts, updates, or deletes
+
+    this is usually instantiated by the model instance to be modified
+
+    unlike the read query object, this is mutable
+    """
+
+    def __init__(self, model, instance=None, batch=None):
+        self.model = model
+        self.column_family_name = self.model.column_family_name()
+        self.instance = instance
+        self.batch = batch
+        pass
+
+    def batch(self, batch_obj):
+        if not isinstance(batch_obj, BatchQuery):
+            raise CQLEngineException('batch_obj must be a BatchQuery instance')
+        self.batch = batch_obj
+        return self
+
+    def save(self):
         """
         Creates / updates a row.
         This is a blind insert call.
-        All validation and cleaning needs to happen 
+        All validation and cleaning needs to happen
         prior to calling this.
         """
-        assert type(instance) == self.model
+        if self.instance is None:
+            raise CQLEngineException("DML Query intance attribute is None")
+        assert type(self.instance) == self.model
 
         #organize data
         value_pairs = []
-        values = instance.as_dict()
+        values = self.instance.as_dict()
 
         #get defined fields and their column names
         for name, col in self.model._columns.items():
@@ -476,63 +661,114 @@ class QuerySet(object):
 
         #construct query string
         field_names = zip(*value_pairs)[0]
+        field_ids = {n:uuid1().hex for n in field_names}
         field_values = dict(value_pairs)
-        qs = ["INSERT INTO {}".format(self.column_family_name)]
-        qs += ["({})".format(', '.join(field_names))]
-        qs += ['VALUES']
-        qs += ["({})".format(', '.join([':'+f for f in field_names]))]
+        query_values = {field_ids[n]:field_values[n] for n in field_names}
+
+        qs = []
+        if self.instance._can_update():
+            qs += ["UPDATE {}".format(self.column_family_name)]
+            qs += ["SET"]
+
+            set_statements = []
+            #get defined fields and their column names
+            for name, col in self.model._columns.items():
+                if not col.is_primary_key:
+                    val = values.get(name)
+                    if val is None: continue
+                    if isinstance(col, BaseContainerColumn):
+                        #remove value from query values, the column will handle it
+                        query_values.pop(field_ids.get(name), None)
+
+                        val_mgr = self.instance._values[name]
+                        set_statements += col.get_update_statement(val, val_mgr.previous_value, query_values)
+                        pass
+                    else:
+                        set_statements += ['"{}" = :{}'.format(col.db_field_name, field_ids[col.db_field_name])]
+            qs += [', '.join(set_statements)]
+
+            qs += ['WHERE']
+
+            where_statements = []
+            for name, col in self.model._primary_keys.items():
+                where_statements += ['"{}" = :{}'.format(col.db_field_name, field_ids[col.db_field_name])]
+
+            qs += [' AND '.join(where_statements)]
+
+            # clear the qs if there are not set statements
+            if not set_statements: qs = []
+
+        else:
+            qs += ["INSERT INTO {}".format(self.column_family_name)]
+            qs += ["({})".format(', '.join(['"{}"'.format(f) for f in field_names]))]
+            qs += ['VALUES']
+            qs += ["({})".format(', '.join([':'+field_ids[f] for f in field_names]))]
+
         qs = ' '.join(qs)
 
-        with connection_manager() as con:
-            con.execute(qs, field_values)
+        # skip query execution if it's empty
+        # caused by pointless update queries
+        if qs:
+            if self.batch:
+                self.batch.add_query(qs, query_values)
+            else:
+                with connection_manager() as con:
+                    con.execute(qs, query_values)
+
+
+        # delete nulled columns and removed map keys
+        qs = ['DELETE']
+        query_values = {}
+
+        del_statements = []
+        for k,v in self.instance._values.items():
+            col = v.column
+            if v.deleted:
+                del_statements += ['"{}"'.format(col.db_field_name)]
+            elif isinstance(col, Map):
+                del_statements += col.get_delete_statement(v.value, v.previous_value, query_values)
+
+        if del_statements:
+            qs += [', '.join(del_statements)]
 
-        #delete deleted / nulled columns
-        deleted = [k for k,v in instance._values.items() if v.deleted]
-        if deleted:
-            del_fields = [self.model._columns[f] for f in deleted]
-            del_fields = [f.db_field_name for f in del_fields if not f.primary_key]
-            pks = self.model._primary_keys
-            qs = ['DELETE {}'.format(', '.join(del_fields))]
             qs += ['FROM {}'.format(self.column_family_name)]
+
             qs += ['WHERE']
-            eq = lambda col: '{0} = :{0}'.format(v.column.db_field_name)
-            qs += [' AND '.join([eq(f) for f in pks.values()])]
+            where_statements = []
+            for name, col in self.model._primary_keys.items():
+                field_id = uuid1().hex
+                query_values[field_id] = field_values[name]
+                where_statements += ['"{}" = :{}'.format(col.db_field_name, field_id)]
+            qs += [' AND '.join(where_statements)]
+
             qs = ' '.join(qs)
 
-            pk_dict = dict([(v.db_field_name, getattr(instance, k)) for k,v in pks.items()])
+            if self.batch:
+                self.batch.add_query(qs, query_values)
+            else:
+                with connection_manager() as con:
+                    con.execute(qs, query_values)
 
-            with connection_manager() as con:
-                con.execute(qs, pk_dict)
-            
-
-    def create(self, **kwargs):
-        return self.model(**kwargs).save()
-
-    #----delete---
-    def delete(self, columns=[]):
-        """
-        Deletes the contents of a query
-        """
-        #validate where clause
-        partition_key = self.model._primary_keys.values()[0]
-        if not any([c.column == partition_key for c in self._where]):
-            raise QueryException("The partition key must be defined on delete queries")
-        qs = ['DELETE FROM {}'.format(self.column_family_name)]
-        qs += ['WHERE {}'.format(self._where_clause())]
-        qs = ' '.join(qs)
-
-        with connection_manager() as con:
-            con.execute(qs, self._where_values())
-
-
-    def delete_instance(self, instance):
+    def delete(self):
         """ Deletes one instance """
-        pk_name = self.model._pk_name
+        if self.instance is None:
+            raise CQLEngineException("DML Query intance attribute is None")
+        field_values = {}
         qs = ['DELETE FROM {}'.format(self.column_family_name)]
-        qs += ['WHERE {0}=:{0}'.format(pk_name)]
+        qs += ['WHERE']
+        where_statements = []
+        for name, col in self.model._primary_keys.items():
+            field_id = uuid1().hex
+            field_values[field_id] = col.to_database(getattr(self.instance, name))
+            where_statements += ['"{}" = :{}'.format(col.db_field_name, field_id)]
+
+        qs += [' AND '.join(where_statements)]
         qs = ' '.join(qs)
 
-        with connection_manager() as con:
-            con.execute(qs, {pk_name:instance.pk})
+        if self.batch:
+            self.batch.add_query(qs, field_values)
+        else:
+            with connection_manager() as con:
+                con.execute(qs, field_values)
 
 
