@@ -1,6 +1,11 @@
+from datetime import datetime
+import time
+from uuid import uuid1, uuid4
+
 from cqlengine.tests.base import BaseCassEngTestCase
 
 from cqlengine.exceptions import ModelException
+from cqlengine import functions
 from cqlengine.management import create_table
 from cqlengine.management import delete_table
 from cqlengine.models import Model
@@ -21,6 +26,12 @@ class IndexedTestModel(Model):
     expected_result = columns.Integer()
     test_result = columns.Integer(index=True)
 
+class TestMultiClusteringModel(Model):
+    one = columns.Integer(primary_key=True)
+    two = columns.Integer(primary_key=True)
+    three = columns.Integer(primary_key=True)
+
+
 class TestQuerySetOperation(BaseCassEngTestCase):
 
     def test_query_filter_parsing(self):
@@ -55,12 +66,12 @@ class TestQuerySetOperation(BaseCassEngTestCase):
         query1 = TestModel.objects(test_id=5)
         ids = [o.identifier for o in query1._where]
         where = query1._where_clause()
-        assert where == 'test_id = :{}'.format(*ids)
+        assert where == '"test_id" = :{}'.format(*ids)
 
         query2 = query1.filter(expected_result__gte=1)
         ids = [o.identifier for o in query2._where]
         where = query2._where_clause()
-        assert where == 'test_id = :{} AND expected_result >= :{}'.format(*ids)
+        assert where == '"test_id" = :{} AND "expected_result" >= :{}'.format(*ids)
 
 
     def test_querystring_generation(self):
@@ -110,6 +121,7 @@ class BaseQuerySetUsage(BaseCassEngTestCase):
         delete_table(IndexedTestModel)
         create_table(TestModel)
         create_table(IndexedTestModel)
+        create_table(TestMultiClusteringModel)
 
         TestModel.objects.create(test_id=0, attempt_id=0, description='try1', expected_result=5, test_result=30)
         TestModel.objects.create(test_id=0, attempt_id=1, description='try2', expected_result=10, test_result=30)
@@ -146,6 +158,7 @@ class BaseQuerySetUsage(BaseCassEngTestCase):
         super(BaseQuerySetUsage, cls).tearDownClass()
         delete_table(TestModel)
         delete_table(IndexedTestModel)
+        delete_table(TestMultiClusteringModel)
 
 class TestQuerySetCountSelectionAndIteration(BaseQuerySetUsage):
 
@@ -165,7 +178,7 @@ class TestQuerySetCountSelectionAndIteration(BaseQuerySetUsage):
             compare_set.remove(val)
         assert len(compare_set) == 0
 
-        q = TestModel.objects(attempt_id=3)
+        q = TestModel.objects(attempt_id=3).allow_filtering()
         assert len(q) == 3
         #tuple of expected test_id, expected_result values
         compare_set = set([(0,20), (1,20), (2,75)])
@@ -174,7 +187,7 @@ class TestQuerySetCountSelectionAndIteration(BaseQuerySetUsage):
             assert val in compare_set
             compare_set.remove(val)
         assert len(compare_set) == 0
-        
+
     def test_multiple_iterations_work_properly(self):
         """ Tests that iterating over a query set more than once works """
         q = TestModel.objects(test_id=0)
@@ -241,6 +254,9 @@ class TestQuerySetCountSelectionAndIteration(BaseQuerySetUsage):
         with self.assertRaises(TestModel.MultipleObjectsReturned):
             TestModel.objects.get(test_id=1)
 
+    def test_allow_filtering_flag(self):
+        """
+        """
 
 class TestQuerySetOrdering(BaseQuerySetUsage):
 
@@ -269,6 +285,23 @@ class TestQuerySetOrdering(BaseQuerySetUsage):
         with self.assertRaises(query.QueryException):
             q = IndexedTestModel.objects(test_id=0).order_by('attempt_id')
 
+    def test_ordering_on_multiple_clustering_columns(self):
+        TestMultiClusteringModel.create(one=1, two=1, three=4)
+        TestMultiClusteringModel.create(one=1, two=1, three=2)
+        TestMultiClusteringModel.create(one=1, two=1, three=5)
+        TestMultiClusteringModel.create(one=1, two=1, three=1)
+        TestMultiClusteringModel.create(one=1, two=1, three=3)
+
+        results = TestMultiClusteringModel.objects.filter(one=1, two=1).order_by('-two', '-three')
+        assert [r.three for r in results] == [5, 4, 3, 2, 1]
+
+        results = TestMultiClusteringModel.objects.filter(one=1, two=1).order_by('two', 'three')
+        assert [r.three for r in results] == [1, 2, 3, 4, 5]
+
+        results = TestMultiClusteringModel.objects.filter(one=1, two=1).order_by('two').order_by('three')
+        assert [r.three for r in results] == [1, 2, 3, 4, 5]
+
+
 class TestQuerySetSlicing(BaseQuerySetUsage):
 
     def test_out_of_range_index_raises_error(self):
@@ -336,12 +369,12 @@ class TestQuerySetDelete(BaseQuerySetUsage):
         TestModel.objects.create(test_id=3, attempt_id=1, description='try10', expected_result=60, test_result=40)
         TestModel.objects.create(test_id=3, attempt_id=2, description='try11', expected_result=70, test_result=45)
         TestModel.objects.create(test_id=3, attempt_id=3, description='try12', expected_result=75, test_result=45)
-        
+
         assert TestModel.objects.count() == 16
         assert TestModel.objects(test_id=3).count() == 4
-        
+
         TestModel.objects(test_id=3).delete()
-        
+
         assert TestModel.objects.count() == 12
         assert TestModel.objects(test_id=3).count() == 0
 
@@ -356,7 +389,7 @@ class TestQuerySetDelete(BaseQuerySetUsage):
             TestModel.objects(attempt_id=0).delete()
 
 class TestQuerySetConnectionHandling(BaseQuerySetUsage):
-    
+
     def test_conn_is_returned_after_filling_cache(self):
         """
         Tests that the queryset returns it's connection after it's fetched all of it's results
@@ -368,10 +401,10 @@ class TestQuerySetConnectionHandling(BaseQuerySetUsage):
             val = t.attempt_id, t.expected_result
             assert val in compare_set
             compare_set.remove(val)
-            
+
         assert q._con is None
         assert q._cur is None
-        
+
     def test_conn_is_returned_after_queryset_is_garbage_collected(self):
         """ Tests that the connection is returned to the connection pool after the queryset is gc'd """
         from cqlengine.connection import ConnectionPool
@@ -379,12 +412,60 @@ class TestQuerySetConnectionHandling(BaseQuerySetUsage):
         q = TestModel.objects(test_id=0)
         v = q[0]
         assert ConnectionPool._queue.qsize() == 0
-        
+
         del q
         assert ConnectionPool._queue.qsize() == 1
 
+class TimeUUIDQueryModel(Model):
+    partition       = columns.UUID(primary_key=True)
+    time            = columns.TimeUUID(primary_key=True)
+    data            = columns.Text(required=False)
 
+class TestMinMaxTimeUUIDFunctions(BaseCassEngTestCase):
 
+    @classmethod
+    def setUpClass(cls):
+        super(TestMinMaxTimeUUIDFunctions, cls).setUpClass()
+        create_table(TimeUUIDQueryModel)
+
+    @classmethod
+    def tearDownClass(cls):
+        super(TestMinMaxTimeUUIDFunctions, cls).tearDownClass()
+        delete_table(TimeUUIDQueryModel)
+
+    def test_success_case(self):
+        """ Test that the min and max time uuid functions work as expected """
+        pk = uuid4()
+        TimeUUIDQueryModel.create(partition=pk, time=uuid1(), data='1')
+        time.sleep(0.2)
+        TimeUUIDQueryModel.create(partition=pk, time=uuid1(), data='2')
+        time.sleep(0.2)
+        midpoint = datetime.utcnow()
+        time.sleep(0.2)
+        TimeUUIDQueryModel.create(partition=pk, time=uuid1(), data='3')
+        time.sleep(0.2)
+        TimeUUIDQueryModel.create(partition=pk, time=uuid1(), data='4')
+        time.sleep(0.2)
+
+        q = TimeUUIDQueryModel.filter(partition=pk, time__lte=functions.MaxTimeUUID(midpoint))
+        q = [d for d in q]
+        assert len(q) == 2
+        datas = [d.data for d in q]
+        assert  '1' in datas
+        assert  '2' in datas
+
+        q = TimeUUIDQueryModel.filter(partition=pk, time__gte=functions.MinTimeUUID(midpoint))
+        assert len(q) == 2
+        datas = [d.data for d in q]
+        assert  '3' in datas
+        assert  '4' in datas
+
+
+class TestInOperator(BaseQuerySetUsage):
+
+    def test_success_case(self):
+        q = TestModel.filter(test_id__in=[0,1])
+        assert q.count() == 8
 
 
 
